@@ -2,7 +2,7 @@
 
   error.c -
 
-  $Author: usa $
+  $Author: nagachika $
   created at: Mon Aug  9 16:11:34 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -25,8 +25,6 @@
 #include <unistd.h>
 #endif
 
-#define numberof(array) (int)(sizeof(array) / sizeof((array)[0]))
-
 #ifndef EXIT_SUCCESS
 #define EXIT_SUCCESS 0
 #endif
@@ -39,6 +37,10 @@
 #define WEXITSTATUS(status) (status)
 #endif
 
+VALUE rb_eEAGAIN;
+VALUE rb_eEWOULDBLOCK;
+VALUE rb_eEINPROGRESS;
+
 extern const char ruby_description[];
 
 static const char REPORTBUG_MSG[] =
@@ -267,6 +269,29 @@ rb_warn_m(int argc, VALUE *argv, VALUE exc)
     return Qnil;
 }
 
+#define MAX_BUG_REPORTERS 0x100
+
+static struct bug_reporters {
+    void (*func)(FILE *out, void *data);
+    void *data;
+} bug_reporters[MAX_BUG_REPORTERS];
+
+static int bug_reporters_size;
+
+int
+rb_bug_reporter_add(void (*func)(FILE *, void *), void *data)
+{
+    struct bug_reporters *reporter;
+    if (bug_reporters_size >= MAX_BUG_REPORTERS) {
+	return 0; /* failed to register */
+    }
+    reporter = &bug_reporters[bug_reporters_size++];
+    reporter->func = func;
+    reporter->data = data;
+
+    return 1;
+}
+
 static void
 report_bug(const char *file, int line, const char *fmt, va_list args)
 {
@@ -284,9 +309,16 @@ report_bug(const char *file, int line, const char *fmt, va_list args)
 	snprintf(buf, 256, "\n%s\n\n", ruby_description);
 	fputs(buf, out);
 
-
 	rb_vm_bugreport();
 
+	/* call additional bug reporters */
+	{
+	    int i;
+	    for (i=0; i<bug_reporters_size; i++) {
+		struct bug_reporters *reporter = &bug_reporters[i];
+		(*reporter->func)(out, reporter->data);
+	    }
+	}
 	fprintf(out, REPORTBUG_MSG);
     }
 }
@@ -535,7 +567,7 @@ VALUE rb_eSystemCallError;
 VALUE rb_mErrno;
 static VALUE rb_eNOERROR;
 
-#undef rb_exc_new2
+#undef rb_exc_new_cstr
 
 VALUE
 rb_exc_new(VALUE etype, const char *ptr, long len)
@@ -544,13 +576,13 @@ rb_exc_new(VALUE etype, const char *ptr, long len)
 }
 
 VALUE
-rb_exc_new2(VALUE etype, const char *s)
+rb_exc_new_cstr(VALUE etype, const char *s)
 {
     return rb_exc_new(etype, s, strlen(s));
 }
 
 VALUE
-rb_exc_new3(VALUE etype, VALUE str)
+rb_exc_new_str(VALUE etype, VALUE str)
 {
     StringValue(str);
     return rb_funcall(etype, rb_intern("new"), 1, str);
@@ -614,11 +646,9 @@ static VALUE
 exc_to_s(VALUE exc)
 {
     VALUE mesg = rb_attr_get(exc, rb_intern("mesg"));
-    VALUE r = Qnil;
 
     if (NIL_P(mesg)) return rb_class_name(CLASS_OF(exc));
-    r = rb_String(mesg);
-    return r;
+    return rb_String(mesg);
 }
 
 /*
@@ -711,6 +741,30 @@ exc_backtrace(VALUE exc)
     return obj;
 }
 
+/*
+ *  call-seq:
+ *     exception.backtrace_locations    -> array
+ *
+ *  Returns any backtrace associated with the exception. This method is
+ *  similar to Exception#backtrace, but the backtrace is an array of
+ *   Thread::Backtrace::Location.
+ *
+ *  Now, this method is not affected by Exception#set_backtrace().
+ */
+static VALUE
+exc_backtrace_locations(VALUE exc)
+{
+    ID bt_locations;
+    VALUE obj;
+
+    CONST_ID(bt_locations, "bt_locations");
+    obj = rb_attr_get(exc, bt_locations);
+    if (!NIL_P(obj)) {
+	obj = rb_backtrace_to_location_ary(obj);
+    }
+    return obj;
+}
+
 VALUE
 rb_check_backtrace(VALUE bt)
 {
@@ -724,7 +778,8 @@ rb_check_backtrace(VALUE bt)
 	    rb_raise(rb_eTypeError, err);
 	}
 	for (i=0;i<RARRAY_LEN(bt);i++) {
-	    if (!RB_TYPE_P(RARRAY_PTR(bt)[i], T_STRING)) {
+	    VALUE e = RARRAY_AREF(bt, i);
+	    if (!RB_TYPE_P(e, T_STRING)) {
 		rb_raise(rb_eTypeError, err);
 	    }
 	}
@@ -754,6 +809,14 @@ rb_exc_set_backtrace(VALUE exc, VALUE bt)
     return exc_set_backtrace(exc, bt);
 }
 
+VALUE
+exc_cause(VALUE exc)
+{
+    ID id_cause;
+    CONST_ID(id_cause, "cause");
+    return rb_attr_get(exc, id_cause);
+}
+
 static VALUE
 try_convert_to_exception(VALUE obj)
 {
@@ -963,24 +1026,6 @@ name_err_name(VALUE self)
     return rb_attr_get(self, rb_intern("name"));
 }
 
-/*
- * call-seq:
- *  name_error.to_s   -> string
- *
- * Produce a nicely-formatted string representing the +NameError+.
- */
-
-static VALUE
-name_err_to_s(VALUE exc)
-{
-    VALUE mesg = rb_attr_get(exc, rb_intern("mesg"));
-    VALUE str = mesg;
-
-    if (NIL_P(mesg)) return rb_class_name(CLASS_OF(exc));
-    StringValue(str);
-    return str;
-}
-
 /*
  * call-seq:
  *   NoMethodError.new(msg, name [, args])  -> no_method_error
@@ -1025,6 +1070,7 @@ static const rb_data_type_t name_err_mesg_data_type = {
 	name_err_mesg_free,
 	name_err_mesg_memsize,
     },
+    NULL, NULL, RUBY_TYPED_FREE_IMMEDIATELY
 };
 
 /* :nodoc: */
@@ -1188,6 +1234,24 @@ set_syserr(int n, const char *name)
 
     if (!st_lookup(syserr_tbl, n, &error)) {
 	error = rb_define_class_under(rb_mErrno, name, rb_eSystemCallError);
+
+	/* capture nonblock errnos for WaitReadable/WaitWritable subclasses */
+	switch (n) {
+	    case EAGAIN:
+	        rb_eEAGAIN = error;
+
+#if EAGAIN != EWOULDBLOCK
+                break;
+	    case EWOULDBLOCK:
+#endif
+
+		rb_eEWOULDBLOCK = error;
+		break;
+	    case EINPROGRESS:
+		rb_eEINPROGRESS = error;
+		break;
+	}
+
 	rb_define_const(error, "Errno", INT2NUM(n));
 	st_add_direct(syserr_tbl, n, error);
     }
@@ -1229,12 +1293,12 @@ syserr_initialize(int argc, VALUE *argv, VALUE self)
     char *strerror();
 #endif
     const char *err;
-    VALUE mesg, error;
+    VALUE mesg, error, func;
     VALUE klass = rb_obj_class(self);
 
     if (klass == rb_eSystemCallError) {
 	st_data_t data = (st_data_t)klass;
-	rb_scan_args(argc, argv, "11", &mesg, &error);
+	rb_scan_args(argc, argv, "12", &mesg, &error, &func);
 	if (argc == 1 && FIXNUM_P(mesg)) {
 	    error = mesg; mesg = Qnil;
 	}
@@ -1244,11 +1308,11 @@ syserr_initialize(int argc, VALUE *argv, VALUE self)
 	    if (!RB_TYPE_P(self, T_OBJECT)) { /* insurance to avoid type crash */
 		rb_raise(rb_eTypeError, "invalid instance type");
 	    }
-	    RBASIC(self)->klass = klass;
+	    RBASIC_SET_CLASS(self, klass);
 	}
     }
     else {
-	rb_scan_args(argc, argv, "01", &mesg);
+	rb_scan_args(argc, argv, "02", &mesg, &func);
 	error = rb_const_get(klass, rb_intern("Errno"));
     }
     if (!NIL_P(error)) err = strerror(NUM2INT(error));
@@ -1258,7 +1322,10 @@ syserr_initialize(int argc, VALUE *argv, VALUE self)
 	VALUE str = StringValue(mesg);
 	rb_encoding *me = rb_enc_get(mesg);
 
-	mesg = rb_sprintf("%s - %"PRIsVALUE, err, mesg);
+	if (NIL_P(func))
+	    mesg = rb_sprintf("%s - %"PRIsVALUE, err, mesg);
+	else
+	    mesg = rb_sprintf("%s @ %"PRIsVALUE" - %"PRIsVALUE, err, func, mesg);
 	if (le != me && rb_enc_asciicompat(me)) {
 	    le = me;
 	}/* else assume err is non ASCII string. */
@@ -1569,14 +1636,14 @@ syserr_eqq(VALUE self, VALUE exc)
  *
  *     foo = "bar"
  *     proc = Proc.new do
- *       $SAFE = 4
- *       foo.gsub! "a", "*"
+ *       $SAFE = 3
+ *       foo.untaint
  *     end
  *     proc.call
  *
  *  <em>raises the exception:</em>
  *
- *     SecurityError: Insecure: can't modify string
+ *     SecurityError: Insecure: Insecure operation `untaint' at level 3
  */
 
 /*
@@ -1711,7 +1778,9 @@ Init_Exception(void)
     rb_define_method(rb_eException, "message", exc_message, 0);
     rb_define_method(rb_eException, "inspect", exc_inspect, 0);
     rb_define_method(rb_eException, "backtrace", exc_backtrace, 0);
+    rb_define_method(rb_eException, "backtrace_locations", exc_backtrace_locations, 0);
     rb_define_method(rb_eException, "set_backtrace", exc_set_backtrace, 1);
+    rb_define_method(rb_eException, "cause", exc_cause, 0);
 
     rb_eSystemExit  = rb_define_class("SystemExit", rb_eException);
     rb_define_method(rb_eSystemExit, "initialize", exit_initialize, -1);
@@ -1733,6 +1802,7 @@ Init_Exception(void)
     rb_eSyntaxError = rb_define_class("SyntaxError", rb_eScriptError);
 
     rb_eLoadError   = rb_define_class("LoadError", rb_eScriptError);
+    /* the path failed to load */
     rb_attr(rb_eLoadError, rb_intern("path"), 1, 0, Qfalse);
 
     rb_eNotImpError = rb_define_class("NotImplementedError", rb_eScriptError);
@@ -1740,7 +1810,6 @@ Init_Exception(void)
     rb_eNameError     = rb_define_class("NameError", rb_eStandardError);
     rb_define_method(rb_eNameError, "initialize", name_err_initialize, -1);
     rb_define_method(rb_eNameError, "name", name_err_name, 0);
-    rb_define_method(rb_eNameError, "to_s", name_err_to_s, 0);
     rb_cNameErrorMesg = rb_define_class_under(rb_eNameError, "message", rb_cData);
     rb_define_singleton_method(rb_cNameErrorMesg, "!", rb_name_err_mesg_new, NAME_ERR_MESG_COUNT);
     rb_define_method(rb_cNameErrorMesg, "==", name_err_mesg_equal, 1);
@@ -1912,6 +1981,34 @@ rb_sys_fail_str(VALUE mesg)
     rb_exc_raise(make_errno_exc_str(mesg));
 }
 
+#ifdef RUBY_FUNCTION_NAME_STRING
+void
+rb_sys_fail_path_in(const char *func_name, VALUE path)
+{
+    int n = errno;
+
+    errno = 0;
+    rb_syserr_fail_path_in(func_name, n, path);
+}
+
+void
+rb_syserr_fail_path_in(const char *func_name, int n, VALUE path)
+{
+    VALUE args[2];
+
+    if (!path) path = Qnil;
+    if (n == 0) {
+	const char *s = !NIL_P(path) ? RSTRING_PTR(path) : "";
+	if (!func_name) func_name = "(null)";
+	rb_bug("rb_sys_fail_path_in(%s, %s) - errno == 0",
+	       func_name, s);
+    }
+    args[0] = path;
+    args[1] = rb_str_new_cstr(func_name);
+    rb_exc_raise(rb_class_new_instance(2, args, get_syserr(n)));
+}
+#endif
+
 void
 rb_mod_sys_fail(VALUE mod, const char *mesg)
 {
@@ -1989,17 +2086,12 @@ rb_check_frozen(VALUE obj)
 void
 rb_error_untrusted(VALUE obj)
 {
-    if (rb_safe_level() >= 4) {
-	rb_raise(rb_eSecurityError, "Insecure: can't modify %s",
-		 rb_obj_classname(obj));
-    }
 }
 
 #undef rb_check_trusted
 void
 rb_check_trusted(VALUE obj)
 {
-    rb_check_trusted_internal(obj);
 }
 
 void
@@ -2007,9 +2099,8 @@ rb_check_copyable(VALUE obj, VALUE orig)
 {
     if (!FL_ABLE(obj)) return;
     rb_check_frozen_internal(obj);
-    rb_check_trusted_internal(obj);
     if (!FL_ABLE(orig)) return;
-    if ((~RBASIC(obj)->flags & RBASIC(orig)->flags) & (FL_UNTRUSTED|FL_TAINT)) {
+    if ((~RBASIC(obj)->flags & RBASIC(orig)->flags) & FL_TAINT) {
 	if (rb_safe_level() > 0) {
 	    rb_raise(rb_eSecurityError, "Insecure: can't modify %"PRIsVALUE,
 		     RBASIC(obj)->klass);
